;; Generated package description from caml.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "caml" "4.10snapshot0.20250227.11" "Caml mode for GNU Emacs" '((emacs "24.4")) :commit "744333dc4c4bd8b93e037efa8f7362b0903b96a2" :authors '(("Jacques Garrigue" . "garrigue@kurims.kyoto-u.ac.jp") ("Ian T Zimmerman" . "itz@rahul.net") ("Damien Doligez" . "damien.doligez@inria.fr")) :maintainer '("Christophe Troestler" . "Christophe.Troestler@umons.ac.be") :keywords '("ocaml") :url "https://github.com/ocaml/caml-mode")
